"""
Pytest fixtures and Konfiguration für questra-authentication Tests
"""

import json
from datetime import datetime, timedelta
from typing import Any
from unittest.mock import MagicMock, Mock

import pytest
from seven2one.questra.authentication.authentication import OidcConfig


@pytest.fixture
def mock_oidc_config() -> OidcConfig:
    """Creates eine Mock OIDC Configuration"""
    return OidcConfig(
        issuer="https://test.example.com",
        authorization_endpoint="https://test.example.com/authorize",
        token_endpoint="https://test.example.com/token",
        userinfo_endpoint="https://test.example.com/userinfo",
        end_session_endpoint="https://test.example.com/logout",
        device_authorization_endpoint="https://test.example.com/device",
    )


@pytest.fixture
def mock_oidc_discovery_response() -> dict[str, str]:
    """Mock response für OIDC Discovery"""
    return {
        "issuer": "https://test.example.com",
        "authorization_endpoint": "https://test.example.com/authorize",
        "token_endpoint": "https://test.example.com/token",
        "userinfo_endpoint": "https://test.example.com/userinfo",
        "end_session_endpoint": "https://test.example.com/logout",
        "device_authorization_endpoint": "https://test.example.com/device",
    }


@pytest.fixture
def mock_token_response() -> dict[str, Any]:
    """Mock OAuth2 Token Response"""
    expires_at = (datetime.now() + timedelta(hours=1)).timestamp()
    return {
        "access_token": "test_access_token_12345",
        "token_type": "Bearer",
        "expires_in": 3600,
        "expires_at": expires_at,
        "refresh_token": "test_refresh_token_67890",
    }


@pytest.fixture
def mock_expired_token_response() -> dict[str, Any]:
    """Mock abgelaufener OAuth2 Token Response"""
    expires_at = (datetime.now() - timedelta(hours=1)).timestamp()
    return {
        "access_token": "expired_access_token",
        "token_type": "Bearer",
        "expires_in": 0,
        "expires_at": expires_at,
        "refresh_token": "test_refresh_token",
    }


@pytest.fixture
def mock_device_code_response() -> dict[str, Any]:
    """Mock Device Code Response für interaktive Authentifizierung"""
    return {
        "device_code": "test_device_code_123",
        "user_code": "ABCD-EFGH",
        "verification_uri": "https://test.example.com/device",
        "verification_uri_complete": "https://test.example.com/device?user_code=ABCD-EFGH",
        "expires_in": 600,
        "interval": 5,
    }


@pytest.fixture
def service_credentials() -> dict[str, str]:
    """Service Account Credentials für Tests"""
    return {"username": "test_service_user", "password": "test_service_password"}


@pytest.fixture
def mock_oauth2_session(mock_token_response: dict[str, Any]) -> MagicMock:
    """Mock OAuth2Session"""
    session = MagicMock()
    session.authorized = True
    session.access_token = mock_token_response["access_token"]
    session.token = mock_token_response
    session.fetch_token.return_value = mock_token_response
    session.refresh_token.return_value = mock_token_response
    return session


@pytest.fixture
def mock_requests_get(
    monkeypatch: pytest.MonkeyPatch, mock_oidc_discovery_response: dict[str, str]
) -> Mock:
    """Mock für requests.get für OIDC Discovery"""
    mock_response = Mock()
    mock_response.json.return_value = mock_oidc_discovery_response
    mock_response.status_code = 200
    mock_response.raise_for_status = Mock()

    mock_get = Mock(return_value=mock_response)
    monkeypatch.setattr("requests.get", mock_get)
    return mock_get


@pytest.fixture
def mock_requests_post(
    monkeypatch: pytest.MonkeyPatch, mock_device_code_response: dict[str, Any]
) -> Mock:
    """Mock für requests.post für Device Code Flow"""
    mock_response = Mock()
    mock_response.text = json.dumps(mock_device_code_response)
    mock_response.json.return_value = mock_device_code_response
    mock_response.status_code = 200

    mock_post = Mock(return_value=mock_response)
    monkeypatch.setattr("requests.request", mock_post)
    return mock_post



"""Custom exceptions for questra-authentication."""


class QuestraAuthenticationError(Exception):
    """Base exception for all questra-authentication errors."""

    pass


class AuthenticationError(QuestraAuthenticationError):
    """Raised when authentication fails."""

    pass


class NotAuthenticatedError(AuthenticationError):
    """Raised when operation requires authentication but not authenticated."""

    pass


class SessionNotInitializedError(AuthenticationError):
    """Raised when OAuth2 session is not initialized."""

    pass


class InvalidCredentialsError(AuthenticationError):
    """Raised when credentials are invalid or of wrong type."""

    pass


class OidcDiscoveryError(QuestraAuthenticationError):
    """Raised when OIDC discovery fails."""

    def __init__(
        self,
        message: str,
        urls: list | None = None,
        original_error: Exception | None = None,
    ):
        """
        Initialize OidcDiscoveryError.

        Args:
            message: Error message
            urls: List of URLs that were tried
            original_error: Original exception that caused the error
        """
        super().__init__(message)
        self.urls = urls or []
        self.original_error = original_error

    def __str__(self) -> str:
        """
        String representation of the error.

        Returns:
            str: Error message with additional details
        """
        msg = super().__str__()
        if self.urls:
            msg += f"\nTried URLs: {', '.join(self.urls)}"
        if self.original_error:
            msg += f"\nOriginal error: {str(self.original_error)}"
        return msg


class TokenExpiredError(AuthenticationError):
    """Raised when a token has expired and cannot be refreshed."""

    pass



"""
Integration Tests für QuestraAuthentication without Mocks.

Diese Tests prüfen die tatsächliche Integration zwischen den Komponenten,
um Probleme like falsche Parameter-Namen to erkennen.
"""

from unittest.mock import Mock, patch

from seven2one.questra.authentication import QuestraAuthentication
from seven2one.questra.authentication.authentication import (
    OAuth2ServiceCredential,
    OidcConfig,
)


class TestQuestraAuthenticationIntegration:
    """Integration Tests die echte Klassen without vollständige Mocks verwenden."""

    @patch("seven2one.questra.authentication.authentication.OAuth2Session")
    @patch("requests.get")
    def test_init_creates_oauth2_authentication_with_correct_parameters(
        self,
        mock_requests_get,
        mock_session_class,
        mock_oidc_discovery_response,
        mock_token_response,
    ):
        """
        Test: QuestraAuthentication initialisiert OAuth2Authentication
        with korrekten Parameter-Namen.

        Dieser Test verwendet die echte OAuth2Authentication-Klasse
        (nicht gemockt) um sicherzustellen, dass Parameter-Namen korrekt sind.
        """
        # Mock OIDC Discovery
        mock_response = Mock()
        mock_response.json.return_value = mock_oidc_discovery_response
        mock_response.raise_for_status = Mock()
        mock_requests_get.return_value = mock_response

        # Mock OAuth2Session
        mock_session = Mock()
        mock_session.fetch_token.return_value = mock_token_response
        mock_session_class.return_value = mock_session

        # Dies sollte without TypeError funktionieren
        client = QuestraAuthentication(
            url="https://test.example.com",
            username="testuser",
            password="testpass",
        )

        # Prüfe dass OAuth2Authentication korrekt initialisiert wurde
        assert client._oauth_client is not None
        assert client._oauth_client.client_id == "Questra"
        assert isinstance(client._oauth_client.credentials, OAuth2ServiceCredential)
        assert isinstance(client._oauth_client.oidc_config, OidcConfig)

    @patch("seven2one.questra.authentication.authentication.OAuth2Session")
    @patch("requests.get")
    def test_oauth2_authentication_parameter_name_validation(
        self,
        mock_requests_get,
        mock_session_class,
        mock_oidc_discovery_response,
        mock_token_response,
    ):
        """
        Test: OAuth2Authentication akzeptiert only 'oidc_config', nicht 'oidcConfig'.

        Dies ist ein Regressions-Test für das CamelCase-Problem.
        """
        # Mock OIDC Discovery
        mock_response = Mock()
        mock_response.json.return_value = mock_oidc_discovery_response
        mock_response.raise_for_status = Mock()
        mock_requests_get.return_value = mock_response

        # Mock OAuth2Session
        mock_session = Mock()
        mock_session.fetch_token.return_value = mock_token_response
        mock_session_class.return_value = mock_session

        # Test sollte durchlaufen without TypeError
        try:
            client = QuestraAuthentication(
                url="https://test.example.com",
                username="testuser",
                password="testpass",
            )
            # If wir hier ankommen, war die Initialisierung erfolgreich
            assert client._oauth_client is not None
        except TypeError as e:
            # Falls ein TypeError auftritt, sollte es NICHT "oidcConfig" erwähnen
            assert "oidcConfig" not in str(e), (
                "OAuth2Authentication was with falschem Parameter-Namen aufgerufen: "
                f"{e}"
            )

    @patch("seven2one.questra.authentication.authentication.OAuth2Session")
    @patch("requests.get")
    def test_full_authentication_flow(
        self,
        mock_requests_get,
        mock_session_class,
        mock_oidc_discovery_response,
        mock_token_response,
    ):
        """
        Test: Vollständiger Authentifizierungs-Flow of QuestraAuthentication
        bis OAuth2Authentication.
        """
        # Mock OIDC Discovery
        mock_response = Mock()
        mock_response.json.return_value = mock_oidc_discovery_response
        mock_response.raise_for_status = Mock()
        mock_requests_get.return_value = mock_response

        # Mock OAuth2Session
        mock_session = Mock()
        mock_session.access_token = mock_token_response["access_token"]
        mock_session.fetch_token.return_value = mock_token_response
        mock_session_class.return_value = mock_session

        # Initialisierung and Authentifizierung
        client = QuestraAuthentication(
            url="https://test.example.com",
            username="testuser",
            password="testpass",
        )

        # Prüfe dass Authentifizierung erfolgreich war
        assert client.is_authenticated()
        assert client._oauth_client is not None
        assert client._oauth_client.authenticated

        # Token abrufen
        token = client.get_access_token()
        assert token == mock_token_response["access_token"]

    @patch("seven2one.questra.authentication.authentication.OAuth2Session")
    @patch("requests.get")
    def test_oidc_config_passed_correctly(
        self,
        mock_requests_get,
        mock_session_class,
        mock_oidc_discovery_response,
        mock_token_response,
    ):
        """
        Test: OidcConfig wird korrekt of QuestraAuthentication
        to OAuth2Authentication übergeben.
        """
        # Mock OIDC Discovery
        mock_response = Mock()
        mock_response.json.return_value = mock_oidc_discovery_response
        mock_response.raise_for_status = Mock()
        mock_requests_get.return_value = mock_response

        # Mock OAuth2Session
        mock_session = Mock()
        mock_session.fetch_token.return_value = mock_token_response
        mock_session_class.return_value = mock_session

        # Initialisierung
        client = QuestraAuthentication(
            url="https://test.example.com",
            username="testuser",
            password="testpass",
        )

        # Prüfe dass oidc_config korrekt übergeben wurde
        assert client._oauth_client is not None
        assert client._oauth_client.oidc_config is not None
        assert client._oauth_client.oidc_config == client._oidc_config
        assert (
            client._oauth_client.oidc_config.token_endpoint
            == mock_oidc_discovery_response["token_endpoint"]
        )

    @patch("seven2one.questra.authentication.authentication.OAuth2Session")
    @patch("requests.get")
    def test_credentials_passed_correctly(
        self,
        mock_requests_get,
        mock_session_class,
        mock_oidc_discovery_response,
        mock_token_response,
    ):
        """
        Test: Credentials werden korrekt of QuestraAuthentication
        to OAuth2Authentication übergeben.
        """
        # Mock OIDC Discovery
        mock_response = Mock()
        mock_response.json.return_value = mock_oidc_discovery_response
        mock_response.raise_for_status = Mock()
        mock_requests_get.return_value = mock_response

        # Mock OAuth2Session
        mock_session = Mock()
        mock_session.fetch_token.return_value = mock_token_response
        mock_session_class.return_value = mock_session

        # Initialisierung with spezifischen Credentials
        test_username = "integration_test_user"
        test_password = "integration_test_pass"

        client = QuestraAuthentication(
            url="https://test.example.com",
            username=test_username,
            password=test_password,
        )

        # Prüfe dass Credentials korrekt übergeben wurden
        assert client._oauth_client is not None
        assert isinstance(client._oauth_client.credentials, OAuth2ServiceCredential)
        assert client._oauth_client.credentials.username == test_username
        assert client._oauth_client.credentials.password == test_password



"""
Questra Authentication - OAuth2 client for Questra API
"""

from .authentication import (
    ClientInterface,
    OAuth2Authentication,
    OAuth2InteractiveUserCredential,
    OAuth2ServiceCredential,
    OidcConfig,
    OidcDiscoveryClient,
)
from .exceptions import (
    AuthenticationError,
    InvalidCredentialsError,
    NotAuthenticatedError,
    OidcDiscoveryError,
    QuestraAuthenticationError,
    SessionNotInitializedError,
    TokenExpiredError,
)
from .questra_authentication import QuestraAuthentication

__all__ = [
    "QuestraAuthentication",
    "OAuth2Authentication",
    "OAuth2ServiceCredential",
    "OAuth2InteractiveUserCredential",
    "OidcConfig",
    "OidcDiscoveryClient",
    "ClientInterface",
    "QuestraAuthenticationError",
    "AuthenticationError",
    "NotAuthenticatedError",
    "SessionNotInitializedError",
    "InvalidCredentialsError",
    "OidcDiscoveryError",
    "TokenExpiredError",
]



"""
QuestraAuthentication - Simplified interface for Questra API authentication
"""

from .authentication import (
    OAuth2Authentication,
    OAuth2InteractiveUserCredential,
    OAuth2ServiceCredential,
    OidcConfig,
    OidcDiscoveryClient,
)


class QuestraAuthentication:
    """
    Main client for Questra API access with automatic OAuth2 authentication.

    Supports two authentication modes:
    1. Service Account (username/password)
    2. Interactive User (Device Code Flow)

    The access token is automatically renewed when it expires.
    """

    def __init__(
        self,
        url: str | list[str],
        client_id: str = "Questra",
        username: str | None = None,
        password: str | None = None,
        interactive: bool = False,
        scope: str | None = None,
        oidc_discovery_paths: list[str] | None = None,
    ):
        """
        Initialize the QuestraAuthentication.

        Args:
            url: Base URL of the Identity Provider or List of URLs
            client_id: OAuth2 Client ID (default: "Questra")
            username: Username for Service Account authentication
            password: Password for Service Account authentication
            interactive: True for interactive Device Code Flow,
                False for Service Account
            scope: Optional OAuth2 Scopes
            oidc_discovery_paths: Optional List of discovery paths
                (e.g. ['/application/o/techstack', '/application/o/questra'])

        Examples:
            Service Account:
            ```python
            client = QuestraAuthentication(
                url="https://authentik.dev.example.com",
                username="ServiceUser",
                password="secret_password",
                oidc_discovery_paths=[
                    "/application/o/techstack",
                    "/application/o/questra",
                ],
            )
            ```

            Interactive:
            ```python
            client = QuestraAuthentication(
                url="https://authentik.dev.example.com",
                interactive=True,
                oidc_discovery_paths=[
                    "/application/o/techstack",
                    "/application/o/questra",
                ],
            )
            ```
        """
        self.url = url
        self.client_id = client_id
        self.username = username
        self.password = password
        self.interactive = interactive
        self.scope = scope or "offline_access"
        self.oidc_discovery_paths = oidc_discovery_paths or ["/application/o/questra"]

        self._oauth_client: OAuth2Authentication | None = None
        self._oidc_config: OidcConfig | None = None

        # Initialize OIDC Discovery and OAuth Client
        self._initialize()

    def _initialize(self):
        """Initialize OIDC Discovery and OAuth2 Client."""
        # OIDC Discovery
        discovery_urls = self._build_discovery_urls()
        oidc_discovery_client = OidcDiscoveryClient(discovery_urls)
        self._oidc_config = oidc_discovery_client.discover()

        # Create OAuth2 Credentials
        if self.interactive:
            credentials = OAuth2InteractiveUserCredential()
        else:
            if not self.username or not self.password:
                raise ValueError(
                    "Username and Password are required for "
                    "Service Account authentication. "
                    "Use interactive=True for interactive login."
                )
            credentials = OAuth2ServiceCredential(
                username=self.username, password=self.password
            )

        # Create OAuth2 Client
        self._oauth_client = OAuth2Authentication(
            client_id=self.client_id,
            credentials=credentials,
            oidc_config=self._oidc_config,
            scope=self.scope,
        )

        # Perform authentication
        self._oauth_client.authenticate()

    def _build_discovery_urls(self) -> str | list[str]:
        """
        Creates Discovery URLs based on Base-URL and discovery paths.

        Returns:
            str | list[str]: String or List of Discovery URLs
        """
        if isinstance(self.url, list):
            # Multiple URLs were passed
            if self.oidc_discovery_paths:
                # Combine all URLs with all discovery paths
                urls = []
                for base_url in self.url:
                    for path in self.oidc_discovery_paths:
                        urls.append(f"{base_url.rstrip('/')}{path}")
                return urls
            else:
                return self.url
        else:
            # Single URL
            if self.oidc_discovery_paths:
                return [
                    f"{self.url.rstrip('/')}{path}"
                    for path in self.oidc_discovery_paths
                ]
            else:
                return self.url

    def get_access_token(self) -> str:
        """
        Returns a valid access token.

        The token is automatically renewed if it has expired.
        This method should be called before each API request.

        Returns:
            str: Valid access token

        Raises:
            Exception: If authentication has failed
        """
        if not self._oauth_client:
            raise Exception("OAuth Client not initialized")

        return self._oauth_client.get_access_token()

    def get_oidc_config(self) -> OidcConfig:
        """
        Returns the OIDC configuration.

        Returns:
            OidcConfig: Object with all endpoints
        """
        if not self._oidc_config:
            raise Exception("OIDC Config not initialized")

        return self._oidc_config

    def is_authenticated(self) -> bool:
        """
        Checks if the client is authenticated.

        Returns:
            True if authenticated, otherwise False
        """
        return self._oauth_client is not None and self._oauth_client.authenticated

    def reauthenticate(self):
        """
        Forces a re-authentication.

        This can be useful if the credentials have changed
        or in case of authentication problems.
        """
        if self._oauth_client:
            self._oauth_client.authenticate()
        else:
            self._initialize()



"""
Unit Tests für OAuth2 Authentication
"""

from datetime import datetime, timedelta
from unittest.mock import MagicMock, Mock, patch

import pytest
from oauthlib.oauth2 import OAuth2Error
from seven2one.questra.authentication.authentication import (
    OAuth2Authentication,
    OAuth2Credential,
    OAuth2InteractiveUserCredential,
    OAuth2ServiceCredential,
)


class TestOAuth2Credentials:
    """Tests für OAuth2 Credential Klassen"""

    def test_create_service_credential(self):
        """Test: OAuth2ServiceCredential kann erstellt werden"""
        cred = OAuth2ServiceCredential(username="testuser", password="testpass")
        assert cred.username == "testuser"
        assert cred.password == "testpass"
        assert isinstance(cred, OAuth2Credential)

    def test_create_interactive_credential(self):
        """Test: OAuth2InteractiveUserCredential kann erstellt werden"""
        cred = OAuth2InteractiveUserCredential()
        assert isinstance(cred, OAuth2Credential)


class TestOAuth2Authentication:
    """Tests für OAuth2Authentication Klasse"""

    def test_init_with_service_credentials(self, mock_oidc_config):
        """Test: Initialisierung with Service Credentials"""
        credentials = OAuth2ServiceCredential("testuser", "testpass")
        auth = OAuth2Authentication(
            client_id="test_client",
            credentials=credentials,
            oidc_config=mock_oidc_config,
        )

        assert auth.client_id == "test_client"
        assert auth.credentials == credentials
        assert auth.oidc_config == mock_oidc_config
        assert not auth.authenticated

    def test_init_with_interactive_credentials(self, mock_oidc_config):
        """Test: Initialisierung with Interactive Credentials"""
        credentials = OAuth2InteractiveUserCredential()
        auth = OAuth2Authentication(
            client_id="test_client",
            credentials=credentials,
            oidc_config=mock_oidc_config,
        )

        assert auth.credentials == credentials
        assert not auth.authenticated

    def test_init_with_invalid_credentials(self, mock_oidc_config):
        """Test: Initialisierung with ungültigen Credentials"""
        # Create a mock that doesn't implement OAuth2Credential interface
        invalid_credentials = Mock(spec=[])  # No methods

        # OAuth2Authentication akzeptiert jetzt alles im __init__,
        # aber authenticate() wird fehlschlagen
        auth = OAuth2Authentication(
            client_id="test_client",
            credentials=invalid_credentials,
            oidc_config=mock_oidc_config,
        )

        # authenticate() sollte fehlschlagen weil create_session fehlt
        with pytest.raises(AttributeError):
            auth.authenticate()

    def test_init_with_scope(self, mock_oidc_config):
        """Test: Initialisierung with Scope"""
        credentials = OAuth2ServiceCredential("testuser", "testpass")
        auth = OAuth2Authentication(
            client_id="test_client",
            credentials=credentials,
            oidc_config=mock_oidc_config,
            scope="openid profile email",
        )

        assert auth.scope == "openid profile email"

    @patch("seven2one.questra.authentication.authentication.OAuth2Session")
    def test_authenticate_service(
        self, mock_session_class, mock_oidc_config, mock_token_response
    ):
        """Test: Authentifizierung with Service Credentials"""
        # Setup mock session
        mock_session = MagicMock()
        mock_session.authorized = True
        mock_session.fetch_token.return_value = mock_token_response
        mock_session_class.return_value = mock_session

        credentials = OAuth2ServiceCredential("testuser", "testpass")
        auth = OAuth2Authentication(
            client_id="test_client",
            credentials=credentials,
            oidc_config=mock_oidc_config,
        )

        auth.authenticate()

        assert auth.authenticated
        assert auth.session == mock_session
        mock_session.fetch_token.assert_called_once()

    @patch("seven2one.questra.authentication.authentication.OAuth2Session")
    def test_authenticate_service_failure(
        self, mock_session_class, mock_oidc_config, mock_token_response
    ):
        """Test: Fehlgeschlagene Service Authentifizierung"""
        mock_session = MagicMock()
        # Simuliere Authentication error durch Exception
        mock_session.fetch_token.side_effect = OAuth2Error("invalid_grant")
        mock_session_class.return_value = mock_session

        credentials = OAuth2ServiceCredential("testuser", "testpass")
        auth = OAuth2Authentication(
            client_id="test_client",
            credentials=credentials,
            oidc_config=mock_oidc_config,
        )

        with pytest.raises(OAuth2Error):
            auth.authenticate()

    @patch("seven2one.questra.authentication.authentication.OAuth2Session")
    @patch("seven2one.questra.authentication.authentication.requests.post")
    @patch("builtins.print")
    def test_authenticate_interactive(
        self,
        mock_print,
        mock_post,
        mock_session_class,
        mock_oidc_config,
        mock_token_response,
        mock_device_code_response,
    ):
        """Test: Interaktive Authentifizierung with Device Code Flow"""
        # Setup device code response
        mock_device_response = Mock()
        mock_device_response.json.return_value = mock_device_code_response
        mock_post.return_value = mock_device_response

        # Setup session mock
        mock_session = MagicMock()
        mock_session.scope = None
        mock_session.fetch_token.return_value = mock_token_response
        mock_session_class.return_value = mock_session

        credentials = OAuth2InteractiveUserCredential()
        auth = OAuth2Authentication(
            client_id="test_client",
            credentials=credentials,
            oidc_config=mock_oidc_config,
        )

        auth.authenticate()

        assert auth.authenticated
        mock_print.assert_called()
        mock_post.assert_called_once()

    def test_get_access_token_not_authenticated(self, mock_oidc_config):
        """Test: get_access_token without Authentifizierung schlägt fehl"""
        credentials = OAuth2ServiceCredential("testuser", "testpass")
        auth = OAuth2Authentication(
            client_id="test_client",
            credentials=credentials,
            oidc_config=mock_oidc_config,
        )

        with pytest.raises(Exception, match="Not authenticated"):
            auth.get_access_token()

    @patch("seven2one.questra.authentication.authentication.OAuth2Session")
    def test_get_access_token_success(
        self, mock_session_class, mock_oidc_config, mock_token_response
    ):
        """Test: get_access_token gibt gültiges Token zurück"""
        mock_session = MagicMock()
        mock_session.authorized = True
        mock_session.access_token = mock_token_response["access_token"]
        mock_session.fetch_token.return_value = mock_token_response
        mock_session_class.return_value = mock_session

        credentials = OAuth2ServiceCredential("testuser", "testpass")
        auth = OAuth2Authentication(
            client_id="test_client",
            credentials=credentials,
            oidc_config=mock_oidc_config,
        )
        auth.authenticate()

        token = auth.get_access_token()

        assert token == mock_token_response["access_token"]

    def test_is_token_expired_not_expired(self, mock_oidc_config):
        """Test: Token ist nicht abgelaufen"""
        credentials = OAuth2ServiceCredential("testuser", "testpass")
        auth = OAuth2Authentication(
            client_id="test_client",
            credentials=credentials,
            oidc_config=mock_oidc_config,
        )

        # Token läuft in 2 Stunden ab
        auth.token_expires_at = datetime.now() + timedelta(hours=2)

        assert not auth._is_token_expired()

    def test_is_token_expired_expired(self, mock_oidc_config):
        """Test: Token ist abgelaufen"""
        credentials = OAuth2ServiceCredential("testuser", "testpass")
        auth = OAuth2Authentication(
            client_id="test_client",
            credentials=credentials,
            oidc_config=mock_oidc_config,
        )

        # Token ist vor 1 Stunde abgelaufen
        auth.token_expires_at = datetime.now() - timedelta(hours=1)

        assert auth._is_token_expired()

    def test_is_token_expired_within_minimum_lifetime(self, mock_oidc_config):
        """Test: Token innerhalb minimaler Lebensdauer gilt as abgelaufen"""
        credentials = OAuth2ServiceCredential("testuser", "testpass")
        auth = OAuth2Authentication(
            client_id="test_client",
            credentials=credentials,
            oidc_config=mock_oidc_config,
        )

        # Token läuft in 30 Sekunden ab (weniger as minimum)
        auth.token_expires_at = datetime.now() + timedelta(seconds=30)
        auth.minimum_token_lifetime_seconds = 60

        assert auth._is_token_expired()

    @patch("seven2one.questra.authentication.authentication.OAuth2Session")
    def test_refresh_authentication_service(
        self, mock_session_class, mock_oidc_config, mock_token_response
    ):
        """Test: Token Refresh für Service Credentials"""
        mock_session = MagicMock()
        mock_session.authorized = True
        mock_session.access_token = "old_token"

        # Neue Token Response with neuem Token
        new_token_response = mock_token_response.copy()
        new_token_response["access_token"] = "new_access_token"
        new_token_response["expires_at"] = (
            datetime.now() + timedelta(hours=2)
        ).timestamp()

        mock_session.fetch_token.return_value = new_token_response
        mock_session_class.return_value = mock_session

        credentials = OAuth2ServiceCredential("testuser", "testpass")
        auth = OAuth2Authentication(
            client_id="test_client",
            credentials=credentials,
            oidc_config=mock_oidc_config,
        )
        auth.authenticate()

        # Simuliere abgelaufenes Token
        auth.token_expires_at = datetime.now() - timedelta(hours=1)

        # Hole neues Token
        auth.get_access_token()

        # fetch_token zweimal aufgerufen (authenticate, refresh)
        assert mock_session.fetch_token.call_count == 2

    @patch("seven2one.questra.authentication.authentication.OAuth2Session")
    @patch("seven2one.questra.authentication.authentication.requests.post")
    def test_refresh_authentication_interactive(
        self,
        mock_post,
        mock_session_class,
        mock_oidc_config,
        mock_token_response,
        mock_device_code_response,
    ):
        """Test: Token Refresh für Interactive Credentials"""
        # Setup device code mock
        mock_device_response = Mock()
        mock_device_response.json.return_value = mock_device_code_response
        mock_post.return_value = mock_device_response

        mock_session = MagicMock()
        mock_session.scope = None
        mock_session.access_token = "old_token"

        # Token Responses
        mock_session.fetch_token.return_value = mock_token_response
        new_token_response = mock_token_response.copy()
        new_token_response["access_token"] = "refreshed_token"
        new_token_response["expires_at"] = (
            datetime.now() + timedelta(hours=2)
        ).timestamp()
        mock_session.refresh_token.return_value = new_token_response

        mock_session_class.return_value = mock_session

        credentials = OAuth2InteractiveUserCredential()
        auth = OAuth2Authentication(
            client_id="test_client",
            credentials=credentials,
            oidc_config=mock_oidc_config,
        )

        with patch("builtins.print"):
            auth.authenticate()

        # Simuliere abgelaufenes Token
        auth.token_expires_at = datetime.now() - timedelta(hours=1)

        # Hole neues Token (sollte refresh_token aufrufen)
        auth.get_access_token()

        mock_session.refresh_token.assert_called_once()

    def test_process_token_response(self, mock_oidc_config, mock_token_response):
        """Test: Token Response wird korrekt verarbeitet"""
        credentials = OAuth2ServiceCredential("testuser", "testpass")
        auth = OAuth2Authentication(
            client_id="test_client",
            credentials=credentials,
            oidc_config=mock_oidc_config,
        )

        auth._process_token_response(mock_token_response)

        assert isinstance(auth.token_expires_at, datetime)

    @patch("seven2one.questra.authentication.authentication.OAuth2Session")
    def test_authenticate_sets_authenticated_flag(
        self, mock_session_class, mock_oidc_config, mock_token_response
    ):
        """Test: authenticate() setzt authenticated Flag"""
        mock_session = MagicMock()
        mock_session.authorized = True
        mock_session.fetch_token.return_value = mock_token_response
        mock_session_class.return_value = mock_session

        credentials = OAuth2ServiceCredential("testuser", "testpass")
        auth = OAuth2Authentication(
            client_id="test_client",
            credentials=credentials,
            oidc_config=mock_oidc_config,
        )

        assert not auth.authenticated
        auth.authenticate()
        assert auth.authenticated



"""
Unit Tests für OIDC Discovery Client
"""

from unittest.mock import Mock, patch

import pytest
from seven2one.questra.authentication.authentication import (
    OidcConfig,
    OidcDiscoveryClient,
)
from seven2one.questra.authentication.exceptions import OidcDiscoveryError


class TestOidcConfig:
    """Tests für OidcConfig Klasse"""

    def test_create_oidc_config(self):
        """Test: OidcConfig kann erstellt werden"""
        config = OidcConfig(
            issuer="https://test.example.com",
            authorization_endpoint="https://test.example.com/authorize",
            token_endpoint="https://test.example.com/token",
            userinfo_endpoint="https://test.example.com/userinfo",
            end_session_endpoint="https://test.example.com/logout",
            device_authorization_endpoint="https://test.example.com/device",
        )

        assert config.issuer == "https://test.example.com"
        assert config.authorization_endpoint == "https://test.example.com/authorize"
        assert config.token_endpoint == "https://test.example.com/token"
        assert config.userinfo_endpoint == "https://test.example.com/userinfo"
        assert config.end_session_endpoint == "https://test.example.com/logout"
        assert config.device_authorization_endpoint == "https://test.example.com/device"

    def test_to_dict(self, mock_oidc_config):
        """Test: OidcConfig.to_dict() gibt Dictionary zurück"""
        config_dict = mock_oidc_config.to_dict()

        assert isinstance(config_dict, dict)
        assert config_dict["issuer"] == "https://test.example.com"
        assert config_dict["token_endpoint"] == "https://test.example.com/token"

    def test_to_json(self, mock_oidc_config):
        """Test: OidcConfig.to_json() gibt JSON String zurück"""
        config_json = mock_oidc_config.to_json()

        assert isinstance(config_json, str)
        assert "https://test.example.com" in config_json


class TestOidcDiscoveryClient:
    """Tests für OidcDiscoveryClient"""

    def test_init_with_string_url(self):
        """Test: Initialisierung with einzelner URL"""
        client = OidcDiscoveryClient("https://test.example.com")
        assert client.base_url == "https://test.example.com"

    def test_init_with_list_of_urls(self):
        """Test: Initialisierung with Liste of URLs"""
        urls = ["https://test1.example.com", "https://test2.example.com"]
        client = OidcDiscoveryClient(urls)
        assert client.base_url == urls

    @patch("requests.get")
    def test_discover_with_direct_url(self, mock_get, mock_oidc_discovery_response):
        """Test: Discovery with direkter URL funktioniert"""
        mock_response = Mock()
        mock_response.json.return_value = mock_oidc_discovery_response
        mock_response.raise_for_status = Mock()
        mock_get.return_value = mock_response

        client = OidcDiscoveryClient("https://test.example.com")
        config = client.discover()

        assert isinstance(config, OidcConfig)
        assert config.issuer == "https://test.example.com"
        mock_get.assert_called_once_with("https://test.example.com")

    @patch("requests.get")
    def test_discover_with_wellknown_fallback(
        self, mock_get, mock_oidc_discovery_response
    ):
        """Test: Discovery with .well-known Fallback"""
        # Erste Anfrage schlägt fehl
        mock_get.side_effect = [
            Exception("Not found"),
            Mock(
                json=lambda: mock_oidc_discovery_response, raise_for_status=lambda: None
            ),
        ]

        client = OidcDiscoveryClient("https://test.example.com")
        config = client.discover()

        assert isinstance(config, OidcConfig)
        assert mock_get.call_count == 2

    @patch("requests.get")
    def test_discover_with_list_tries_all_urls(
        self, mock_get, mock_oidc_discovery_response
    ):
        """Test: Discovery with Liste probiert alle URLs"""
        # Erste URL schlägt fehl, zweite funktioniert
        mock_get.side_effect = [
            Exception("Not found"),
            Exception("Not found"),
            Mock(
                json=lambda: mock_oidc_discovery_response, raise_for_status=lambda: None
            ),
        ]

        urls = ["https://test1.example.com", "https://test2.example.com"]
        client = OidcDiscoveryClient(urls)
        config = client.discover()

        assert isinstance(config, OidcConfig)
        assert mock_get.call_count >= 2

    @patch("requests.get")
    def test_discover_fails_when_no_valid_url(self, mock_get):
        """Test: Discovery schlägt fehl wenn keine URL funktioniert"""
        mock_get.side_effect = Exception("Not found")

        client = OidcDiscoveryClient("https://test.example.com")

        with pytest.raises(Exception, match="OIDC discovery failed"):
            client.discover()

    @patch("requests.get")
    def test_discover_fails_with_list_when_all_urls_fail(self, mock_get):
        """Test: Discovery with Liste schlägt fehl wenn alle URLs fehlschlagen"""
        mock_get.side_effect = Exception("Not found")

        urls = ["https://test1.example.com", "https://test2.example.com"]
        client = OidcDiscoveryClient(urls)

        with pytest.raises(OidcDiscoveryError, match="No valid configuration found"):
            client.discover()

    @patch("requests.get")
    def test_fetch_raises_http_error(self, mock_get):
        """Test: _fetch gibt HTTP Fehler weiter"""
        mock_response = Mock()
        mock_response.raise_for_status.side_effect = Exception("HTTP Error")
        mock_get.return_value = mock_response

        client = OidcDiscoveryClient("https://test.example.com")

        with pytest.raises(Exception):
            client._fetch("https://test.example.com")

    @patch("requests.get")
    def test_discover_extracts_all_endpoints(
        self, mock_get, mock_oidc_discovery_response
    ):
        """Test: Discovery extrahiert alle Endpoints korrekt"""
        mock_response = Mock()
        mock_response.json.return_value = mock_oidc_discovery_response
        mock_response.raise_for_status = Mock()
        mock_get.return_value = mock_response

        client = OidcDiscoveryClient("https://test.example.com")
        config = client.discover()

        assert config.issuer == mock_oidc_discovery_response["issuer"]
        assert (
            config.authorization_endpoint
            == mock_oidc_discovery_response["authorization_endpoint"]
        )
        assert config.token_endpoint == mock_oidc_discovery_response["token_endpoint"]
        assert (
            config.userinfo_endpoint
            == mock_oidc_discovery_response["userinfo_endpoint"]
        )
        assert (
            config.end_session_endpoint
            == mock_oidc_discovery_response["end_session_endpoint"]
        )
        assert (
            config.device_authorization_endpoint
            == mock_oidc_discovery_response["device_authorization_endpoint"]
        )

    @patch("requests.get")
    def test_discover_fails_with_missing_required_fields(self, mock_get):
        """Test: Discovery schlägt fehl wenn erforderliche Felder fehlen"""
        # Response with fehlendem token_endpoint
        incomplete_response = {
            "issuer": "https://test.example.com",
            "authorization_endpoint": "https://test.example.com/authorize",
            # token_endpoint fehlt!
            "userinfo_endpoint": "https://test.example.com/userinfo",
            "end_session_endpoint": "https://test.example.com/logout",
            "device_authorization_endpoint": "https://test.example.com/device",
        }

        mock_response = Mock()
        mock_response.json.return_value = incomplete_response
        mock_response.raise_for_status = Mock()
        mock_get.return_value = mock_response

        client = OidcDiscoveryClient("https://test.example.com")

        with pytest.raises(OidcDiscoveryError, match="missing required fields"):
            client.discover()



import json
import logging
import time
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from datetime import datetime, timedelta
from typing import Any

import requests
from oauthlib.oauth2 import BackendApplicationClient, DeviceClient, OAuth2Error
from requests_oauthlib import OAuth2Session

from .exceptions import (
    AuthenticationError,
    NotAuthenticatedError,
    OidcDiscoveryError,
    SessionNotInitializedError,
)

logger = logging.getLogger(__name__)


class ClientInterface(ABC):
    """Abstract interface for OAuth2 clients."""

    @abstractmethod
    def get_access_token(self) -> str:
        """
        Get access token.

        Returns:
            str: Access token
        """
        pass


@dataclass(frozen=True)
class OidcConfig:
    """OIDC Configuration containing OAuth2 endpoints."""

    issuer: str
    authorization_endpoint: str
    token_endpoint: str
    userinfo_endpoint: str
    end_session_endpoint: str
    device_authorization_endpoint: str

    def to_dict(self) -> dict[str, str]:
        """
        Convert to dictionary.

        Returns:
            dict[str, str]: Dictionary representation
        """
        return asdict(self)

    def to_json(self) -> str:
        """
        Convert to JSON string.

        Returns:
            str: JSON representation
        """
        return json.dumps(self.to_dict())


class OidcDiscoveryClient:
    def __init__(self, base_url: str | list[str]):
        self.base_url = base_url

    def discover(self) -> OidcConfig:
        config = None
        tried_urls = []

        if isinstance(self.base_url, str):
            urls_to_try = [self.base_url]
        elif isinstance(self.base_url, list):
            urls_to_try = self.base_url
        else:
            raise OidcDiscoveryError("base_url must be string or list of strings")

        for url in urls_to_try:
            tried_urls.append(url)
            try:
                config = self._fetch(url)
                break
            except Exception:
                well_known_url = f"{url}/.well-known/openid-configuration"
                tried_urls.append(well_known_url)
                try:
                    config = self._fetch(well_known_url)
                    break
                except Exception:
                    continue

        if config is None:
            raise OidcDiscoveryError(
                "OIDC discovery failed. No valid configuration found.", urls=tried_urls
            )

        # Validate required fields
        required_fields = [
            "issuer",
            "authorization_endpoint",
            "token_endpoint",
            "userinfo_endpoint",
            "end_session_endpoint",
            "device_authorization_endpoint",
        ]
        missing_fields = [field for field in required_fields if not config.get(field)]
        if missing_fields:
            missing = ", ".join(missing_fields)
            raise OidcDiscoveryError(
                f"OIDC configuration missing required fields: {missing}",
                urls=tried_urls,
            )

        return OidcConfig(
            issuer=config["issuer"],
            authorization_endpoint=config["authorization_endpoint"],
            token_endpoint=config["token_endpoint"],
            userinfo_endpoint=config["userinfo_endpoint"],
            end_session_endpoint=config["end_session_endpoint"],
            device_authorization_endpoint=config["device_authorization_endpoint"],
        )

    def _fetch(self, url: str) -> dict[str, Any]:
        """Fetch OIDC configuration from URL."""
        response = requests.get(url)
        response.raise_for_status()
        return response.json()


class OAuth2Credential(ABC):
    """Abstract base class for OAuth2 credentials."""

    @abstractmethod
    def create_session(
        self, client_id: str, oidc_config: "OidcConfig", scope: str | None = None
    ) -> OAuth2Session:
        """
        Create OAuth2 session for this credential type.

        Returns:
            OAuth2Session: OAuth2 session instance
        """
        pass

    @abstractmethod
    def authenticate(
        self, session: OAuth2Session, client_id: str, oidc_config: "OidcConfig"
    ) -> dict[str, Any]:
        """
        Perform authentication and return token response.

        Returns:
            dict[str, Any]: Token response
        """
        pass

    @abstractmethod
    def refresh(
        self, session: OAuth2Session, client_id: str, oidc_config: "OidcConfig"
    ) -> dict[str, Any]:
        """
        Refresh authentication token.

        Returns:
            dict[str, Any]: Token response
        """
        pass


@dataclass
class OAuth2ServiceCredential(OAuth2Credential):
    """Service account credentials (username/password)."""

    username: str
    password: str

    def create_session(
        self, client_id: str, oidc_config: "OidcConfig", scope: str | None = None
    ) -> OAuth2Session:
        """
        Create OAuth2 session with backend application client.

        Returns:
            OAuth2Session: OAuth2 session instance
        """
        client = BackendApplicationClient(client_id=client_id)
        return OAuth2Session(client=client)

    def authenticate(
        self, session: OAuth2Session, client_id: str, oidc_config: "OidcConfig"
    ) -> dict[str, Any]:
        """
        Authenticate using username/password.

        Returns:
            dict[str, Any]: Token response
        """
        return session.fetch_token(
            token_url=oidc_config.token_endpoint,
            client_id=client_id,
            username=self.username,
            password=self.password,
        )

    def refresh(
        self, session: OAuth2Session, client_id: str, oidc_config: "OidcConfig"
    ) -> dict[str, Any]:
        """
        Refresh token using username/password.

        Returns:
            dict[str, Any]: Token response
        """
        return session.fetch_token(
            token_url=oidc_config.token_endpoint,
            client_id=client_id,
            username=self.username,
            password=self.password,
        )


@dataclass
class OAuth2InteractiveUserCredential(OAuth2Credential):
    """Interactive user credentials (device code flow)."""

    def create_session(
        self, client_id: str, oidc_config: "OidcConfig", scope: str | None = None
    ) -> OAuth2Session:
        """
        Create OAuth2 session with device client.

        Returns:
            OAuth2Session: OAuth2 session instance
        """
        device_client = DeviceClient(client_id=client_id)
        return OAuth2Session(
            client=device_client,
            auto_refresh_url=oidc_config.token_endpoint,
            scope=scope,
        )

    def authenticate(
        self, session: OAuth2Session, client_id: str, oidc_config: "OidcConfig"
    ) -> dict[str, Any]:
        """
        Authenticate using device code flow.

        Returns:
            dict[str, Any]: Token response
        """
        device_code_response = self._get_device_code(
            client_id, oidc_config, session.scope
        )

        logger.info(
            "Please go to %s and authorize access.",
            device_code_response["verification_uri"],
        )
        uri = device_code_response["verification_uri"]
        print(f"Please go to {uri} and authorize access.")

        while True:
            try:
                token_response = session.fetch_token(
                    token_url=oidc_config.token_endpoint,
                    device_code=device_code_response["device_code"],
                    include_client_id=True,
                )
                return token_response
            except OAuth2Error as e:
                if e.error == "authorization_pending":
                    time.sleep(device_code_response["interval"])
                else:
                    logger.error("Authorization failed: %s", e.error)
                    raise AuthenticationError(f"Authorization failed: {e.error}") from e

    def refresh(
        self, session: OAuth2Session, client_id: str, oidc_config: "OidcConfig"
    ) -> dict[str, Any]:
        """
        Refresh token.

        Returns:
            dict[str, Any]: Token response
        """
        return session.refresh_token(
            token_url=oidc_config.token_endpoint, client_id=client_id
        )

    def _get_device_code(
        self, client_id: str, oidc_config: "OidcConfig", scope: str | None = None
    ) -> dict[str, Any]:
        """
        Get device code from authorization endpoint.

        Returns:
            dict[str, Any]: Device code response
        """
        scope_arg = f"&scope={scope}" if scope else ""
        response = requests.post(
            oidc_config.device_authorization_endpoint,
            data=f"client_id={client_id}{scope_arg}",
            headers={"Content-Type": "application/x-www-form-urlencoded"},
        )
        response.raise_for_status()
        json_response = response.json()

        return {
            "device_code": json_response["device_code"],
            "interval": json_response["interval"],
            "verification_uri": json_response["verification_uri_complete"],
            "expires_in": json_response["expires_in"],
        }


class OAuth2Authentication(ClientInterface):
    """OAuth2 authentication client using Strategy pattern for credentials."""

    def __init__(
        self,
        client_id: str,
        credentials: OAuth2Credential,
        oidc_config: OidcConfig,
        scope: str | None = None,
        minimum_token_lifetime_seconds: int = 60,
    ):
        """
        Initialize OAuth2 authentication client.

        Args:
            client_id: OAuth2 client ID
            credentials: Credential strategy (Service or Interactive)
            oidc_config: OIDC configuration with endpoints
            scope: Optional OAuth2 scopes
            minimum_token_lifetime_seconds: Minimum lifetime before token refresh
        """
        self.client_id = client_id
        self.credentials = credentials
        self.oidc_config = oidc_config
        self.scope = scope
        self.minimum_token_lifetime_seconds = minimum_token_lifetime_seconds

        self.session: OAuth2Session | None = None
        self.authenticated = False
        self.token_expires_at: datetime | None = None

    def authenticate(self) -> None:
        """Perform initial authentication."""
        logger.info("Starting authentication")
        self.session = self.credentials.create_session(
            self.client_id, self.oidc_config, self.scope
        )
        token_response = self.credentials.authenticate(
            self.session, self.client_id, self.oidc_config
        )
        self._process_token_response(token_response)
        self.authenticated = True
        logger.info("Authentication successful")

    def get_access_token(self) -> str:
        """
        Get valid access token, refreshing if necessary.

        Returns:
            str: Valid access token
        """
        if not self.authenticated:
            raise NotAuthenticatedError("Not authenticated. Call authenticate() first")
        if not self.session:
            raise SessionNotInitializedError("Session not initialized")

        self._refresh_if_needed()
        return self.session.access_token

    def _refresh_if_needed(self) -> None:
        """Refresh token if expired or close to expiry."""
        if not self._is_token_expired():
            return

        logger.info("Token expired, refreshing")
        if not self.session:
            raise SessionNotInitializedError("Session not initialized")

        token_response = self.credentials.refresh(
            self.session, self.client_id, self.oidc_config
        )
        self._process_token_response(token_response)
        logger.info("Token refreshed successfully")

    def _process_token_response(self, response: dict[str, Any]) -> None:
        """Process token response and update expiry time."""
        self.token_expires_at = datetime.fromtimestamp(response["expires_at"])

    def _is_token_expired(self) -> bool:
        """
        Check if token is expired or close to expiry.

        Returns:
            bool: True if token is expired or near expiry, False otherwise
        """
        if not self.token_expires_at:
            return True
        return (
            self.token_expires_at
            - timedelta(seconds=self.minimum_token_lifetime_seconds)
        ) < datetime.now()



"""
Unit Tests für QuestraAuthentication
"""

from unittest.mock import MagicMock, patch

import pytest
from seven2one.questra.authentication import QuestraAuthentication
from seven2one.questra.authentication.authentication import OidcConfig


class TestQuestraAuthentication:
    """Tests für QuestraAuthentication Klasse"""

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_init_with_service_credentials(
        self,
        mock_oauth_class,
        mock_discovery_class,
        mock_oidc_config,
        mock_oauth2_session,
    ):
        """Test: Initialisierung with Service Account Credentials"""
        # Setup mocks
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = True
        mock_oauth_class.return_value = mock_oauth

        client = QuestraAuthentication(
            url="https://test.example.com",
            username="testuser",
            password="testpass",
            oidc_discovery_paths=["/application/o/questra"],
        )

        assert client.url == "https://test.example.com"
        assert client.username == "testuser"
        assert client.password == "testpass"
        assert not client.interactive
        assert client.client_id == "Questra"
        mock_oauth.authenticate.assert_called_once()

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_init_with_interactive_mode(
        self, mock_oauth_class, mock_discovery_class, mock_oidc_config
    ):
        """Test: Initialisierung with interaktivem Modus"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = True
        mock_oauth_class.return_value = mock_oauth

        client = QuestraAuthentication(
            url="https://test.example.com",
            interactive=True,
            oidc_discovery_paths=["/application/o/questra"],
        )

        assert client.interactive
        assert client.username is None
        assert client.password is None
        mock_oauth.authenticate.assert_called_once()

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_init_with_custom_client_id(
        self, mock_oauth_class, mock_discovery_class, mock_oidc_config
    ):
        """Test: Initialisierung with benutzerdefinierter Client ID"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = True
        mock_oauth_class.return_value = mock_oauth

        client = QuestraAuthentication(
            url="https://test.example.com",
            client_id="CustomClient",
            username="testuser",
            password="testpass",
        )

        assert client.client_id == "CustomClient"

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_init_with_scope(
        self, mock_oauth_class, mock_discovery_class, mock_oidc_config
    ):
        """Test: Initialisierung with OAuth Scope"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = True
        mock_oauth_class.return_value = mock_oauth

        client = QuestraAuthentication(
            url="https://test.example.com",
            username="testuser",
            password="testpass",
            scope="openid profile email",
        )

        assert client.scope == "openid profile email"

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    def test_init_without_credentials_raises_error(
        self, mock_discovery_class, mock_oidc_config
    ):
        """Test: Initialisierung without Credentials schlägt fehl"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        with pytest.raises(
            ValueError,
            match="Username and Password are required for Service Account authentication. Use interactive=True for interactive login.",
        ):
            QuestraAuthentication(url="https://test.example.com")

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_get_access_token(
        self,
        mock_oauth_class,
        mock_discovery_class,
        mock_oidc_config,
        mock_token_response,
    ):
        """Test: get_access_token gibt Token zurück"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = True
        mock_oauth.get_access_token.return_value = mock_token_response["access_token"]
        mock_oauth_class.return_value = mock_oauth

        client = QuestraAuthentication(
            url="https://test.example.com", username="testuser", password="testpass"
        )

        token = client.get_access_token()

        assert token == mock_token_response["access_token"]
        mock_oauth.get_access_token.assert_called_once()

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_get_access_token_without_oauth_client(
        self, mock_oauth_class, mock_discovery_class, mock_oidc_config
    ):
        """Test: get_access_token without OAuth Client schlägt fehl"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = True
        mock_oauth_class.return_value = mock_oauth

        client = QuestraAuthentication(
            url="https://test.example.com", username="testuser", password="testpass"
        )

        client._oauth_client = None

        with pytest.raises(Exception, match="OAuth Client not initialized"):
            client.get_access_token()

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_get_oidc_config(
        self, mock_oauth_class, mock_discovery_class, mock_oidc_config
    ):
        """Test: get_oidc_config gibt OidcConfig zurück"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = True
        mock_oauth_class.return_value = mock_oauth

        client = QuestraAuthentication(
            url="https://test.example.com", username="testuser", password="testpass"
        )

        config = client.get_oidc_config()

        assert isinstance(config, OidcConfig)
        assert config == mock_oidc_config

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_get_oidc_config_without_config(
        self, mock_oauth_class, mock_discovery_class, mock_oidc_config
    ):
        """Test: get_oidc_config without Config schlägt fehl"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = True
        mock_oauth_class.return_value = mock_oauth

        client = QuestraAuthentication(
            url="https://test.example.com", username="testuser", password="testpass"
        )

        client._oidc_config = None

        with pytest.raises(Exception, match="OIDC Config not initialized"):
            client.get_oidc_config()

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_is_authenticated(
        self, mock_oauth_class, mock_discovery_class, mock_oidc_config
    ):
        """Test: is_authenticated gibt korrekten Status zurück"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = True
        mock_oauth_class.return_value = mock_oauth

        client = QuestraAuthentication(
            url="https://test.example.com", username="testuser", password="testpass"
        )

        assert client.is_authenticated()

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_is_authenticated_false(
        self, mock_oauth_class, mock_discovery_class, mock_oidc_config
    ):
        """Test: is_authenticated gibt False zurück wenn nicht authentifiziert"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = False
        mock_oauth_class.return_value = mock_oauth

        client = QuestraAuthentication(
            url="https://test.example.com", username="testuser", password="testpass"
        )

        assert not client.is_authenticated()

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_reauthenticate(
        self, mock_oauth_class, mock_discovery_class, mock_oidc_config
    ):
        """Test: reauthenticate ruft authenticate erneut auf"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = True
        mock_oauth_class.return_value = mock_oauth

        client = QuestraAuthentication(
            url="https://test.example.com", username="testuser", password="testpass"
        )

        # authenticate was already in __init__ aufgerufen
        initial_call_count = mock_oauth.authenticate.call_count

        client.reauthenticate()

        # authenticate sollte erneut aufgerufen worden sein
        assert mock_oauth.authenticate.call_count == initial_call_count + 1

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_build_discovery_urls_single_url_with_paths(
        self, mock_oauth_class, mock_discovery_class, mock_oidc_config
    ):
        """Test: _build_discovery_urls with einzelner URL and Pfaden"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = True
        mock_oauth_class.return_value = mock_oauth

        client = QuestraAuthentication(
            url="https://test.example.com",
            username="testuser",
            password="testpass",
            oidc_discovery_paths=["/path1", "/path2"],
        )

        urls = client._build_discovery_urls()

        assert isinstance(urls, list)
        assert "https://test.example.com/path1" in urls
        assert "https://test.example.com/path2" in urls

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_build_discovery_urls_multiple_urls_with_paths(
        self, mock_oauth_class, mock_discovery_class, mock_oidc_config
    ):
        """Test: _build_discovery_urls with mehreren URLs and Pfaden"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = True
        mock_oauth_class.return_value = mock_oauth

        client = QuestraAuthentication(
            url=["https://test1.example.com", "https://test2.example.com"],
            username="testuser",
            password="testpass",
            oidc_discovery_paths=["/path1", "/path2"],
        )

        urls = client._build_discovery_urls()

        assert isinstance(urls, list)
        assert len(urls) == 4  # 2 URLs * 2 Pfade
        assert "https://test1.example.com/path1" in urls
        assert "https://test1.example.com/path2" in urls
        assert "https://test2.example.com/path1" in urls
        assert "https://test2.example.com/path2" in urls

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_build_discovery_urls_single_url_without_paths(
        self, mock_oauth_class, mock_discovery_class, mock_oidc_config
    ):
        """Test: _build_discovery_urls with einzelner URL without Pfade verwendet Default"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = True
        mock_oauth_class.return_value = mock_oauth

        # If oidc_discovery_paths=None (nicht angegeben), wird Default verwendet
        client = QuestraAuthentication(
            url="https://test.example.com",
            username="testuser",
            password="testpass",
            # oidc_discovery_paths nicht angegeben
            # -> Default: ['/application/o/questra']
        )

        urls = client._build_discovery_urls()

        # With Default-Pfad
        assert isinstance(urls, list)
        assert "https://test.example.com/application/o/questra" in urls

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_build_discovery_urls_multiple_urls_without_paths(
        self, mock_oauth_class, mock_discovery_class, mock_oidc_config
    ):
        """Test: _build_discovery_urls with mehreren URLs without Pfade verwendet Default"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = True
        mock_oauth_class.return_value = mock_oauth

        urls_list = ["https://test1.example.com", "https://test2.example.com"]
        client = QuestraAuthentication(
            url=urls_list,
            username="testuser",
            password="testpass",
            # oidc_discovery_paths nicht angegeben
            # -> Default: ['/application/o/questra']
        )

        urls = client._build_discovery_urls()

        # With Default-Pfad für jede URL
        assert isinstance(urls, list)
        assert "https://test1.example.com/application/o/questra" in urls
        assert "https://test2.example.com/application/o/questra" in urls

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_build_discovery_urls_strips_trailing_slashes(
        self, mock_oauth_class, mock_discovery_class, mock_oidc_config
    ):
        """Test: _build_discovery_urls entfernt trailing slashes"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = True
        mock_oauth_class.return_value = mock_oauth

        client = QuestraAuthentication(
            url="https://test.example.com/",
            username="testuser",
            password="testpass",
            oidc_discovery_paths=["/path1"],
        )

        urls = client._build_discovery_urls()

        assert "https://test.example.com/path1" in urls
        assert "https://test.example.com//path1" not in urls

    @patch(
        "seven2one.questra.authentication.questra_authentication.OidcDiscoveryClient"
    )
    @patch(
        "seven2one.questra.authentication.questra_authentication.OAuth2Authentication"
    )
    def test_default_oidc_discovery_paths(
        self, mock_oauth_class, mock_discovery_class, mock_oidc_config
    ):
        """Test: Standard OIDC Discovery Pfade werden verwendet"""
        mock_discovery = MagicMock()
        mock_discovery.discover.return_value = mock_oidc_config
        mock_discovery_class.return_value = mock_discovery

        mock_oauth = MagicMock()
        mock_oauth.authenticated = True
        mock_oauth_class.return_value = mock_oauth

        client = QuestraAuthentication(
            url="https://test.example.com",
            username="testuser",
            password="testpass",
            # oidc_discovery_paths nicht angegeben
        )

        assert client.oidc_discovery_paths == ["/application/o/questra"]
